# SPDX-License-Identifier: Apache-2.0

# Copyright 2020 Contributors to OpenLEADR

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from . import service, handler, VTNService
import asyncio
from openleadr import objects, utils, enums
import logging
import sys
from datetime import datetime, timezone
from functools import partial
from dataclasses import asdict
logger = logging.getLogger('openleadr')


@service('EiEvent')
class EventService(VTNService):

    def __init__(self, vtn_id, polling_method='internal', message_queues=None):
        super().__init__(vtn_id)
        self.polling_method = polling_method
        self.message_queues = message_queues
        self.pending_events = {}        # Holds the event callbacks
        self.running_events = {}        # Holds the event callbacks for accepted events

    @handler('oadrRequestEvent')
    async def request_event(self, payload):
        """
        The VEN requests us to send any events we have.
        """
        if self.polling_method == 'external':
            result = self.on_request_event(ven_id=payload['ven_id'])
            if asyncio.iscoroutine(result):
                result = await result
        elif payload['ven_id'] in self.message_queues:
            queue = self.message_queues[payload['ven_id']]
            result = utils.get_next_event_from_deque(queue)
        else:
            return 'oadrResponse', {}

        if result is None:
            return 'oadrResponse', {}
        if isinstance(result, dict) and 'event_descriptor' in result:
            return 'oadrDistributeEvent', {'events': [result]}
        elif isinstance(result, objects.Event):
            return 'oadrDistributeEvent', {'events': [asdict(result)]}

        logger.warning("Could not determine type of message "
                       f"in response to oadrRequestEvent: {result}")
        return 'oadrResponse', result

    def on_request_event(self, ven_id):
        """
        Placeholder for the on_request_event handler.
        """
        logger.warning("You should implement and register your own on_request_event handler "
                       "that returns the next event for a VEN. This handler will receive a "
                       "ven_id as its only argument, and should return None (if no events are "
                       "available), a single Event, or a list of Events.")
        return None

    @handler('oadrCreatedEvent')
    async def created_event(self, payload):
        """
        The VEN informs us that they created an EiEvent.
        """
        ven_id = payload['ven_id']
        if self.polling_method == 'internal':
            for event_response in payload['event_responses']:
                event_id = event_response['event_id']
                opt_type = event_response['opt_type']
                if event_response['event_id'] in self.pending_events:
                    event, callback = self.pending_events.pop(event_id)
                    if isinstance(callback, asyncio.Future):
                        callback.set_result(opt_type)
                    else:
                        result = callback(ven_id=ven_id, event_id=event_id, opt_type=opt_type)
                        if asyncio.iscoroutine(result):
                            result = await result
                    if opt_type == 'optIn':
                        self.running_events[event_id] = (event, callback)
                        self.schedule_event_updates(ven_id, event)
                elif event_response['event_id'] in self.running_events:
                    event, callback = self.running_events.pop(event_id)
                    if isinstance(callback, asyncio.Future):
                        logger.warning(f"Got a second response '{opt_type}' from ven '{ven_id}' "
                                       f"to event '{event_id}', which we cannot use because the "
                                       "callback future you provided was already completed during "
                                       "the first response.")
                    else:
                        result = callback(ven_id=ven_id, event_id=event_id, opt_type=opt_type)
                        if asyncio.iscoroutine(result):
                            result = await result
        else:
            result = self.on_created_event(ven_id=ven_id, event_id=event_id, opt_type=opt_type)
            if asyncio.iscoroutine(result):
                result = await(result)
        return 'oadrResponse', {}

    def on_created_event(self, ven_id, event_id, opt_type):
        """
        Placeholder for the on_created_event handler.
        """
        logger.warning("You should implement and register you own on_created_event handler "
                       "to receive the opt status for an Event that you sent to the VEN. This "
                       "handler will receive a ven_id, event_id and opt_status. "
                       "You don't need to return anything from this handler.")
        return None

    def _update_event_status(self, ven_id, event, event_status):
        """
        Update the event to the given Status.
        """
        event.event_descriptor.event_status = event_status
        if event_status == enums.EVENT_STATUS.CANCELLED:
            event.event_descriptor.modification_number += 1
        self.message_queues[ven_id].append(event)

    def schedule_event_updates(self, ven_id, event):
        """
        Schedules the event updates.
        """
        loop = asyncio.get_event_loop()
        now = datetime.now(timezone.utc)
        active_period = event.active_period

        # Named tasks is only supported in Python 3.8+
        if sys.version_info.minor >= 8:
            named_tasks = True
        else:
            named_tasks = False
            name = {}

        # Schedule status update to 'near' if applicable
        if active_period.ramp_up_period is not None and event.event_descriptor.event_status == 'far':
            ramp_up_start_delay = (active_period.dtstart - active_period.ramp_up_period) - now
            update_coro = partial(self._update_event_status, ven_id, event, 'near')
            if named_tasks:
                name = {'name': f'DelayedCall-EventStatusToNear-{event.event_descriptor.event_id}'}
            loop.create_task(utils.delayed_call(func=update_coro, delay=ramp_up_start_delay), **name)

        # Schedule status update to 'active'
        if event.event_descriptor.event_status in ('near', 'far'):
            active_start_delay = active_period.dtstart - now
            update_coro = partial(self._update_event_status, ven_id, event, 'active')
            if named_tasks:
                name = {'name': f'DelayedCall-EventStatusToActive-{event.event_descriptor.event_id}'}
            loop.create_task(utils.delayed_call(func=update_coro, delay=active_start_delay), **name)

        # Schedule status update to 'completed'
        if event.event_descriptor.event_status in ('near', 'far', 'active'):
            active_end_delay = active_period.dtstart + active_period.duration - now
            update_coro = partial(self._update_event_status, ven_id, event, 'completed')
            if named_tasks:
                name = {'name': f'DelayedCall-EventStatusToActive-{event.event_descriptor.event_id}'}
            loop.create_task(utils.delayed_call(func=update_coro, delay=active_end_delay), **name)



# SPDX-License-Identifier: Apache-2.0

# Copyright 2020 Contributors to OpenLEADR

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
A collection of useful enumerations that you can use to construct or
interpret OpenADR messages. Can also be useful during testing.
"""
from openleadr.objects import Measurement, PowerAttributes


class Enum(type):
    def __getitem__(self, item):
        return getattr(self, item)

    @property
    def members(self):
        return sorted([item for item in list(set(dir(self)) - set(dir(Enum)))
                       if not item.startswith("_")])

    @property
    def values(self):
        return [self[item] for item in self.members]


class EVENT_STATUS(metaclass=Enum):
    NONE = "none"
    FAR = "far"
    NEAR = "near"
    ACTIVE = "active"
    COMPLETED = "completed"
    CANCELLED = "cancelled"


class SIGNAL_TYPE(metaclass=Enum):
    DELTA = "delta"
    LEVEL = "level"
    MULTIPLIER = "multiplier"
    PRICE = "price"
    PRICE_MULTIPLIER = "priceMultiplier"
    PRICE_RELATIVE = "priceRelative"
    SETPOINT = "setpoint"
    X_LOAD_CONTROL_CAPACITY = "x-loadControlCapacity"
    X_LOAD_CONTROL_LEVEL_OFFSET = "x-loadControlLevelOffset"
    X_LOAD_CONTROL_PERCENT_OFFSET = "x-loadControlPorcentOffset"
    X_LOAD_CONTROL_SETPOINT = "x-loadControlSetpoint"


class SIGNAL_NAME(metaclass=Enum):
    SIMPLE = "SIMPLE"
    simple = "simple"
    ELECTRICITY_PRICE = "ELECTRICITY_PRICE"
    ENERGY_PRICE = "ENERGY_PRICE"
    DEMAND_CHARGE = "DEMAND_CHARGE"
    BID_PRICE = "BID_PRICE"
    BID_LOAD = "BID_LOAD"
    BID_ENERGY = "BID_ENERGY"
    CHARGE_STATE = "CHARGE_STATE"
    LOAD_DISPATCH = "LOAD_DISPATCH"
    LOAD_CONTROL = "LOAD_CONTROL"


class SI_SCALE_CODE(metaclass=Enum):
    p = "p"
    n = "n"
    micro = "micro"
    m = "m"
    c = "c"
    d = "d"
    k = "k"
    M = "M"
    G = "G"
    T = "T"
    none = "none"


class OPT(metaclass=Enum):
    OPT_IN = "optIn"
    OPT_OUT = "optOut"


class OPT_REASON(metaclass=Enum):
    ECONOMIC = "economic"
    EMERGENCY = "emergency"
    MUST_RUN = "mustRun"
    NOT_PARTICIPATING = "notParticipating"
    OUTAGE_RUN_STATUS = "outageRunStatus"
    OVERRIDE_STATUS = "overrideStatus"
    PARTICIPATING = "participating"
    X_SCHEDULE = "x-schedule"


class READING_TYPE(metaclass=Enum):
    DIRECT_READ = "Direct Read"
    NET = "Net"
    ALLOCATED = "Allocated"
    ESTIMATED = "Estimated"
    SUMMED = "Summed"
    DERIVED = "Derived"
    MEAN = "Mean"
    PEAK = "Peak"
    HYBRID = "Hybrid"
    CONTRACT = "Contract"
    PROJECTED = "Projected"
    X_RMS = "x-RMS"
    X_NOT_APPLICABLE = "x-notApplicable"


class REPORT_TYPE(metaclass=Enum):
    READING = "reading"
    USAGE = "usage"
    DEMAND = "demand"
    SET_POINT = "setPoint"
    DELTA_USAGE = "deltaUsage"
    DELTA_SET_POINT = "deltaSetPoint"
    DELTA_DEMAND = "deltaDemand"
    BASELINE = "baseline"
    DEVIATION = "deviation"
    AVG_USAGE = "avgUsage"
    AVG_DEMAND = "avgDemand"
    OPERATING_STATE = "operatingState"
    UP_REGULATION_CAPACITY_AVAILABLE = "upRegulationCapacityAvailable"
    DOWN_REGULATION_CAPACITY_AVAILABLE = "downRegulationCapacityAvailable"
    REGULATION_SETPOINT = "regulationSetpoint"
    STORED_ENERGY = "storedEnergy"
    TARGET_ENERGY_STORAGE = "targetEnergyStorage"
    AVAILABLE_ENERGY_STORAGE = "availableEnergyStorage"
    PRICE = "price"
    LEVEL = "level"
    POWER_FACTOR = "powerFactor"
    PERCENT_USAGE = "percentUsage"
    PERCENT_DEMAND = "percentDemand"
    X_RESOURCE_STATUS = "x-resourceStatus"


class REPORT_NAME(metaclass=Enum):
    METADATA_HISTORY_USAGE = "METADATA_HISTORY_USAGE"
    HISTORY_USAGE = "HISTORY_USAGE"
    METADATA_HISTORY_GREENBUTTON = "METADATA_HISTORY_GREENBUTTON"
    HISTORY_GREENBUTTON = "HISTORY_GREENBUTTON"
    METADATA_TELEMETRY_USAGE = "METADATA_TELEMETRY_USAGE"
    TELEMETRY_USAGE = "TELEMETRY_USAGE"
    METADATA_TELEMETRY_STATUS = "METADATA_TELEMETRY_STATUS"
    TELEMETRY_STATUS = "TELEMETRY_STATUS"


class STATUS_CODES(metaclass=Enum):
    OUT_OF_SEQUENCE = 450
    NOT_ALLOWED = 451
    INVALID_ID = 452
    NOT_RECOGNIZED = 453
    INVALID_DATA = 454
    COMPLIANCE_ERROR = 459
    SIGNAL_NOT_SUPPORTED = 460
    REPORT_NOT_SUPPORTED = 461
    TARGET_MISMATCH = 462
    NOT_REGISTERED_OR_AUTHORIZED = 463
    DEPLOYMENT_ERROR_OR_OTHER_ERROR = 469


class SECURITY_LEVEL:
    STANDARD = 'STANDARD'
    HIGH = 'HIGH'


_CURRENCIES = ("AFN", "ALL", "AMD", "ANG", "AOA", "ARS", "AUD", "AWG", "AZN", "BAM",
               "BBD", "BDT", "BGN", "BHD", "BIF", "BMD", "BND", "BOB", "BOV", "BRL",
               "BSD", "BTN", "BWP", "BYR", "BZD", "CAD", "CDF", "CHE", "CHF", "CHW",
               "CLF", "CLP", "CNY", "COP", "COU", "CRC", "CUC", "CUP", "CVE", "CZK",
               "DJF", "DKK", "DOP", "DZD", "EEK", "EGP", "ERN", "ETB", "EUR", "FJD",
               "FKP", "GBP", "GEL", "GHS", "GIP", "GMD", "GNF", "GTQ", "GWP", "GYD",
               "HKD", "HNL", "HRK", "HTG", "HUF", "IDR", "ILS", "INR", "IQD", "IRR",
               "ISK", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRW",
               "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LVL",
               "LYD", "MAD", "MAD", "MDL", "MGA", "MKD", "MMK", "MNT", "MOP", "MRO",
               "MUR", "MVR", "MWK", "MXN", "MXV", "MYR", "MZN", "NAD", "NGN", "NIO",
               "NOK", "NPR", "NZD", "OMR", "PAB", "PEN", "PGK", "PHP", "PKR", "PLN",
               "PYG", "QAR", "RON", "RSD", "RUB", "RWF", "SAR", "SBD", "SCR", "SDG",
               "SEK", "SGD", "SHP", "SLL", "SOS", "SRD", "STD", "SVC", "SYP", "SZL",
               "THB", "TJS", "TMT", "TND", "TOP", "TRY", "TTD", "TWD", "TZS", "UAH",
               "UGX", "USD", "USN", "USS", "UYI", "UYU", "UZS", "VEF", "VND", "VUV",
               "WST", "XAF", "XAG", "XAU", "XBA", "XBB", "XBC", "XBD", "XCD", "XDR",
               "XFU", "XOF", "XPD", "XPF", "XPF", "XPF", "XPT", "XTS", "XXX", "YER",
               "ZAR", "ZMK", "ZWL")

_ACCEPTABLE_UNITS = {'currency': _CURRENCIES,
                     'currencyPerKW': _CURRENCIES,
                     'currencyPerKWh': _CURRENCIES,
                     'currencyPerThm': _CURRENCIES,
                     'current': ('A',),
                     'energyApparent': ('VAh',),
                     'energyReactive': ('VARh',),
                     'energyReal': ('Wh',),
                     'frequency': ('Hz',),
                     'powerApparent': ('VA',),
                     'powerReactive': ('VAR',),
                     'powerReal': ('W',),
                     'pulseCount': ('count',),
                     'temperature': ('celsius', 'fahrenheit'),
                     'Therm': ('thm',),
                     'voltage': ('V',)}

_MEASUREMENT_DESCRIPTIONS = {'currency': 'currency',
                             'currencyPerKW': 'currencyPerKW',
                             'currencyPerKWh': 'currencyPerKWh',
                             'currencyPerThm': 'currency',
                             'current': 'Current',
                             'energyApparent': 'ApparentEnergy',
                             'energyReactive': 'ReactiveEnergy',
                             'energyReal': 'RealEnergy',
                             'frequency': 'Frequency',
                             'powerApparent': 'ApparentPower',
                             'powerReactive': 'ReactivePower',
                             'powerReal': 'RealPower',
                             'pulseCount': 'pulse count',
                             'temperature': 'temperature',
                             'Therm': 'Therm',
                             'voltage': 'Voltage'}

_MEASUREMENT_NAMESPACES = {'currency': 'oadr',
                           'currencyPerWK': 'oadr',
                           'currencyPerKWh': 'oadr',
                           'currencyPerThm': 'oadr',
                           'current': 'oadr',
                           'energyApparent': 'power',
                           'energyReactive': 'power',
                           'energyReal': 'power',
                           'frequency': 'oadr',
                           'powerApparent': 'power',
                           'powerReactive': 'power',
                           'powerReal': 'power',
                           'pulseCount': 'oadr',
                           'temperature': 'oadr',
                           'Therm': 'oadr',
                           'voltage': 'power',
                           'customUnit': 'oadr'}


class MEASUREMENTS(metaclass=Enum):
    VOLTAGE = Measurement(name='voltage',
                          description=_MEASUREMENT_DESCRIPTIONS['voltage'],
                          unit=_ACCEPTABLE_UNITS['voltage'][0],
                          acceptable_units=_ACCEPTABLE_UNITS['voltage'],
                          scale='none')
    CURRENT = Measurement(name='current',
                          description=_MEASUREMENT_DESCRIPTIONS['current'],
                          unit=_ACCEPTABLE_UNITS['current'][0],
                          acceptable_units=_ACCEPTABLE_UNITS['current'],
                          scale='none')
    ENERGY_REAL = Measurement(name='energyReal',
                              description=_MEASUREMENT_DESCRIPTIONS['energyReal'],
                              unit=_ACCEPTABLE_UNITS['energyReal'][0],
                              acceptable_units=_ACCEPTABLE_UNITS['energyReal'],
                              scale='none')
    REAL_ENERGY = Measurement(name='energyReal',
                              description=_MEASUREMENT_DESCRIPTIONS['energyReal'],
                              unit=_ACCEPTABLE_UNITS['energyReal'][0],
                              acceptable_units=_ACCEPTABLE_UNITS['energyReal'],
                              scale='none')
    ACTIVE_ENERGY = Measurement(name='energyReal',
                                description=_MEASUREMENT_DESCRIPTIONS['energyReal'],
                                unit=_ACCEPTABLE_UNITS['energyReal'][0],
                                acceptable_units=_ACCEPTABLE_UNITS['energyReal'],
                                scale='none')
    ENERGY_REACTIVE = Measurement(name='energyReactive',
                                  description=_MEASUREMENT_DESCRIPTIONS['energyReactive'],
                                  unit=_ACCEPTABLE_UNITS['energyReactive'][0],
                                  acceptable_units=_ACCEPTABLE_UNITS['energyReactive'],
                                  scale='none')
    REACTIVE_ENERGY = Measurement(name='energyReactive',
                                  description=_MEASUREMENT_DESCRIPTIONS['energyReactive'],
                                  unit=_ACCEPTABLE_UNITS['energyReactive'][0],
                                  acceptable_units=_ACCEPTABLE_UNITS['energyReactive'],
                                  scale='none')
    ENERGY_APPARENT = Measurement(name='energyApparent',
                                  description=_MEASUREMENT_DESCRIPTIONS['energyApparent'],
                                  unit=_ACCEPTABLE_UNITS['energyApparent'][0],
                                  acceptable_units=_ACCEPTABLE_UNITS['energyApparent'],
                                  scale='none')
    APPARENT_ENERGY = Measurement(name='energyApparent',
                                  description=_MEASUREMENT_DESCRIPTIONS['energyApparent'],
                                  unit=_ACCEPTABLE_UNITS['energyApparent'][0],
                                  acceptable_units=_ACCEPTABLE_UNITS['energyApparent'],
                                  scale='none')
    ACTIVE_POWER = Measurement(name='powerReal',
                               description=_MEASUREMENT_DESCRIPTIONS['powerReal'],
                               unit=_ACCEPTABLE_UNITS['powerReal'][0],
                               acceptable_units=_ACCEPTABLE_UNITS['powerReal'],
                               scale='none',
                               power_attributes=PowerAttributes(hertz=50,
                                                                voltage=230,
                                                                ac=True))
    REAL_POWER = Measurement(name='powerReal',
                             description=_MEASUREMENT_DESCRIPTIONS['powerReal'],
                             unit=_ACCEPTABLE_UNITS['powerReal'][0],
                             acceptable_units=_ACCEPTABLE_UNITS['powerReal'],
                             scale='none',
                             power_attributes=PowerAttributes(hertz=50,
                                                              voltage=230,
                                                              ac=True))
    POWER_REAL = Measurement(name='powerReal',
                             description=_MEASUREMENT_DESCRIPTIONS['powerReal'],
                             unit=_ACCEPTABLE_UNITS['powerReal'][0],
                             acceptable_units=_ACCEPTABLE_UNITS['powerReal'],
                             scale='none',
                             power_attributes=PowerAttributes(hertz=50,
                                                              voltage=230,
                                                              ac=True))
    REACTIVE_POWER = Measurement(name='powerReactive',
                                 description=_MEASUREMENT_DESCRIPTIONS['powerReactive'],
                                 unit=_ACCEPTABLE_UNITS['powerReactive'][0],
                                 acceptable_units=_ACCEPTABLE_UNITS['powerReactive'],
                                 scale='none',
                                 power_attributes=PowerAttributes(hertz=50,
                                                                  voltage=230,
                                                                  ac=True))
    POWER_REACTIVE = Measurement(name='powerReactive',
                                 description=_MEASUREMENT_DESCRIPTIONS['powerReactive'],
                                 unit=_ACCEPTABLE_UNITS['powerReactive'][0],
                                 acceptable_units=_ACCEPTABLE_UNITS['powerReactive'],
                                 scale='none',
                                 power_attributes=PowerAttributes(hertz=50,
                                                                  voltage=230,
                                                                  ac=True))
    APPARENT_POWER = Measurement(name='powerApparent',
                                 description=_MEASUREMENT_DESCRIPTIONS['powerApparent'],
                                 unit=_ACCEPTABLE_UNITS['powerApparent'][0],
                                 acceptable_units=_ACCEPTABLE_UNITS['powerApparent'],
                                 scale='none',
                                 power_attributes=PowerAttributes(hertz=50,
                                                                  voltage=230,
                                                                  ac=True))
    POWER_APPARENT = Measurement(name='powerApparent',
                                 description=_MEASUREMENT_DESCRIPTIONS['powerApparent'],
                                 unit=_ACCEPTABLE_UNITS['powerApparent'][0],
                                 acceptable_units=_ACCEPTABLE_UNITS['powerApparent'],
                                 scale='none',
                                 power_attributes=PowerAttributes(hertz=50,
                                                                  voltage=230,
                                                                  ac=True))
    FREQUENCY = Measurement(name='frequency',
                            description=_MEASUREMENT_DESCRIPTIONS['frequency'],
                            unit=_ACCEPTABLE_UNITS['frequency'][0],
                            acceptable_units=_ACCEPTABLE_UNITS['frequency'],
                            scale='none')
    PULSE_COUNT = Measurement(name='pulseCount',
                              description=_MEASUREMENT_DESCRIPTIONS['pulseCount'],
                              unit=_ACCEPTABLE_UNITS['pulseCount'][0],
                              acceptable_units=_ACCEPTABLE_UNITS['pulseCount'],
                              pulse_factor=1000)
    TEMPERATURE = Measurement(name='temperature',
                              description=_MEASUREMENT_DESCRIPTIONS['temperature'],
                              unit=_ACCEPTABLE_UNITS['temperature'][0],
                              acceptable_units=_ACCEPTABLE_UNITS['temperature'],
                              scale='none')
    THERM = Measurement(name='Therm',
                        description=_MEASUREMENT_DESCRIPTIONS['Therm'],
                        unit=_ACCEPTABLE_UNITS['Therm'][0],
                        acceptable_units=_ACCEPTABLE_UNITS['Therm'],
                        scale='none')
    CURRENCY = Measurement(name='currency',
                           description=_MEASUREMENT_DESCRIPTIONS['currency'],
                           unit=_CURRENCIES[0],
                           acceptable_units=_CURRENCIES,
                           scale='none')
    CURRENCY_PER_KW = Measurement(name='currencyPerKW',
                                  description=_MEASUREMENT_DESCRIPTIONS['currencyPerKW'],
                                  unit=_CURRENCIES[0],
                                  acceptable_units=_CURRENCIES,
                                  scale='none')
    CURRENCY_PER_KWH = Measurement(name='currencyPerKWh',
                                   description=_MEASUREMENT_DESCRIPTIONS['currencyPerKWh'],
                                   unit=_CURRENCIES[0],
                                   acceptable_units=_CURRENCIES,
                                   scale='none')
    CURRENCY_PER_THM = Measurement(name='currencyPerThm',
                                   description=_MEASUREMENT_DESCRIPTIONS['currencyPerThm'],
                                   unit=_CURRENCIES[0],
                                   acceptable_units=_CURRENCIES,
                                   scale='none')



# SPDX-License-Identifier: Apache-2.0

# Copyright 2020 Contributors to OpenLEADR

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from . import service, handler, VTNService
from asyncio import iscoroutine
import logging
logger = logging.getLogger('openleadr')

# ╔══════════════════════════════════════════════════════════════════════════╗
# ║                           REGISTRATION SERVICE                           ║
# ╚══════════════════════════════════════════════════════════════════════════╝
# ┌──────────────────────────────────────────────────────────────────────────┐
# │ The VEN can explore some information about the VTN:                      │
# │                                                                          │
# │ ┌────┐                                                            ┌────┐ │
# │ │VEN │                                                            │VTN │ │
# │ └─┬──┘                                                            └─┬──┘ │
# │   │─────────────────────oadrQueryRegistration()────────────────────▶│    │
# │   │                                                                 │    │
# │   │◀ ─ ─ ─ ─ ─ ─ oadrCreatedPartyRegistration(VTN Info)─ ─ ─ ─ ─ ─ ─│    │
# │   │                                                                 │    │
# │                                                                          │
# └──────────────────────────────────────────────────────────────────────────┘
# ┌──────────────────────────────────────────────────────────────────────────┐
# │ The VEN can then go on and register with the VTN                         │
# │                                                                          │
# │ ┌────┐                                                            ┌────┐ │
# │ │VEN │                                                            │VTN │ │
# │ └─┬──┘                                                            └─┬──┘ │
# │   │───────────────oadrCreatePartyRegistration(VEN Info)────────────▶│    │
# │   │                                                                 │    │
# │   │◀ ─ ─ oadrCreatedPartyRegistration(VTN Info, registrationID)─ ─ ─│    │
# │   │                                                                 │    │
# │                                                                          │
# └──────────────────────────────────────────────────────────────────────────┘
# ┌──────────────────────────────────────────────────────────────────────────┐
# │ The VEN can also choose to cancel the registration                       │
# │                                                                          │
# │ ┌────┐                                                            ┌────┐ │
# │ │VEN │                                                            │VTN │ │
# │ └─┬──┘                                                            └─┬──┘ │
# │   │──────────oadrCancelPartyRegistration(registrationID)───────────▶│    │
# │   │                                                                 │    │
# │   │◀ ─ ─ ─ ─ ─ ─ ─ ─oadrCanceledPartyRegistration()─ ─ ─ ─ ─ ─ ─ ─ ─│    │
# │   │                                                                 │    │
# │                                                                          │
# └──────────────────────────────────────────────────────────────────────────┘


@service('EiRegisterParty')
class RegistrationService(VTNService):

    def __init__(self, vtn_id, poll_freq):
        super().__init__(vtn_id)
        self.poll_freq = poll_freq

    @handler('oadrQueryRegistration')
    async def query_registration(self, payload):
        """
        Return the profiles we support.
        """
        if hasattr(self, 'on_query_registration'):
            result = self.on_query_registration(payload)
            if iscoroutine(result):
                result = await result
            return result

        # If you don't provide a default handler, just give out the info
        response_payload = {'request_id': payload['request_id'],
                            'profiles': [{'profile_name': '2.0b',
                                          'transports': [{'transport_name': 'simpleHttp'}]}],
                            'requested_oadr_poll_freq': self.poll_freq}
        return 'oadrCreatedPartyRegistration', response_payload

    @handler('oadrCreatePartyRegistration')
    async def create_party_registration(self, payload):
        """
        Handle the registration of a VEN party.
        """
        result = self.on_create_party_registration(payload)
        if iscoroutine(result):
            result = await result

        if result is not False and result is not None:
            if len(result) != 2:
                logger.error("Your on_create_party_registration handler should return either "
                             "'False' (if the client is rejected) or a (ven_id, registration_id) "
                             "tuple. Will REJECT the client for now.")
                response_payload = {}
            else:
                ven_id, registration_id = result
                transports = [{'transport_name': payload['transport_name']}]
                response_payload = {'ven_id': result[0],
                                    'registration_id': result[1],
                                    'profiles': [{'profile_name': payload['profile_name'],
                                                  'transports': transports}],
                                    'requested_oadr_poll_freq': self.poll_freq}
        else:
            transports = [{'transport_name': payload['transport_name']}]
            response_payload = {'profiles': [{'profile_name': payload['profile_name'],
                                              'transports': transports}],
                                'requested_oadr_poll_freq': self.poll_freq}
        return 'oadrCreatedPartyRegistration', response_payload

    def on_create_party_registration(self, payload):
        """
        Placeholder for the on_create_party_registration handler
        """
        logger.warning("You should implement and register your own on_create_party_registration "
                       "handler if you want VENs to be able to connect to you. This handler will "
                       "receive a registration request and should return either 'False' (if the "
                       "registration is denied) or a (ven_id, registration_id) tuple if the "
                       "registration is accepted.")
        return False

    @handler('oadrCancelPartyRegistration')
    async def cancel_party_registration(self, payload):
        """
        Cancel the registration of a party.
        """
        result = self.on_cancel_party_registration(payload)
        if iscoroutine(result):
            result = await result
        return result

    def on_cancel_party_registration(self, ven_id):
        """
        Placeholder for the on_cancel_party_registration handler.
        """
        logger.warning("You should implement and register your own on_cancel_party_registration "
                       "handler that allown VENs to deregister from your VTN. This will receive a "
                       "ven_id as its argument. You don't need to return anything.")
        return None



# SPDX-License-Identifier: Apache-2.0

# Copyright 2020 Contributors to OpenLEADR

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from asyncio import iscoroutine
from http import HTTPStatus
import logging

from aiohttp import web
from lxml.etree import XMLSyntaxError
from signxml.exceptions import InvalidSignature

from .. import errors
from ..enums import STATUS_CODES
from ..messaging import parse_message, validate_xml_schema, authenticate_message
from ..utils import generate_id, get_cert_fingerprint_from_request

from dataclasses import is_dataclass, asdict

logger = logging.getLogger('openleadr')


class VTNService:
    def __init__(self, vtn_id):
        self.vtn_id = vtn_id
        self.handlers = {}
        for method in [getattr(self, attr) for attr in dir(self) if callable(getattr(self, attr))]:
            if hasattr(method, '__message_type__'):
                self.handlers[method.__message_type__] = method

    async def handler(self, request):
        """
        Handle all incoming POST requests.
        """
        content = await request.read()
        try:
            # Validate the message to the XML Schema
            message_tree = validate_xml_schema(content)

            # Parse the message to a type and payload dict
            message_type, message_payload = parse_message(content)

            # Authenticate the message
            if request.secure and 'ven_id' in message_payload:
                await authenticate_message(request, message_tree, message_payload,
                                           self.fingerprint_lookup)

            # Pass the message off to the handler and get the response type and payload
            try:
                # Add the request fingerprint to the message so that the handler can check for it.
                if request.secure and message_type == 'oadrCreatePartyRegistration':
                    message_payload['fingerprint'] = get_cert_fingerprint_from_request(request)
                response_type, response_payload = await self.handle_message(message_type,
                                                                            message_payload)
            except Exception as err:
                logger.error("An exception occurred during the execution of your handler: "
                             f"{err.__class__.__name__}: {err}")
                raise err

            if 'response' not in response_payload:
                response_payload['response'] = {'response_code': 200,
                                                'response_description': 'OK',
                                                'request_id': message_payload.get('request_id')}
            response_payload['vtn_id'] = self.vtn_id
            if 'ven_id' not in response_payload:
                response_payload['ven_id'] = message_payload.get('ven_id')
        except errors.ProtocolError as err:
            # In case of an OpenADR error, return a valid OpenADR message
            response_type, response_payload = self.error_response(message_type,
                                                                  err.response_code,
                                                                  err.response_description)
            msg = self._create_message(response_type, **response_payload)
            response = web.Response(text=msg,
                                    status=HTTPStatus.OK,
                                    content_type='application/xml')
        except errors.HTTPError as err:
            # If we throw a http-related error, deal with it here
            response = web.Response(text=err.response_description,
                                    status=err.response_code)
        except XMLSyntaxError as err:
            logger.warning(f"XML schema validation of incoming message failed: {err}.")
            response = web.Response(text=f'XML failed validation: {err}',
                                    status=HTTPStatus.BAD_REQUEST)
        except errors.FingerprintMismatch as err:
            logger.warning(err)
            response = web.Response(text=str(err),
                                    status=HTTPStatus.FORBIDDEN)
        except InvalidSignature:
            logger.warning("Incoming message had invalid signature, ignoring.")
            response = web.Response(text='Invalid Signature',
                                    status=HTTPStatus.FORBIDDEN)
        except Exception as err:
            # In case of some other error, return a HTTP 500
            logger.error(f"The VTN server encountered an error: {err.__class__.__name__}: {err}")
            response = web.Response(status=HTTPStatus.INTERNAL_SERVER_ERROR)
        else:
            # We've successfully handled this message
            msg = self._create_message(response_type, **response_payload)
            response = web.Response(text=msg,
                                    status=HTTPStatus.OK,
                                    content_type='application/xml')
        return response

    async def handle_message(self, message_type, message_payload):
        if message_type in self.handlers:
            handler = self.handlers[message_type]
            result = handler(message_payload)
            if iscoroutine(result):
                result = await result
            if result is not None:
                response_type, response_payload = result
                if is_dataclass(response_payload):
                    response_payload = asdict(response_payload)
                elif response_payload is None:
                    response_payload = {}
            else:
                response_type, response_payload = 'oadrResponse', {}

            response_payload['vtn_id'] = self.vtn_id
            if 'ven_id' in message_payload:
                response_payload['ven_id'] = message_payload['ven_id']

            response_payload['response'] = {'request_id': message_payload.get('request_id', None),
                                            'response_code': 200,
                                            'response_description': 'OK'}
            response_payload['request_id'] = generate_id()

        else:
            response_type, response_payload = self.error_response('oadrResponse',
                                                                  STATUS_CODES.COMPLIANCE_ERROR,
                                                                  "A message of type "
                                                                  f"{message_type} should not be "
                                                                  "sent to this endpoint")
        logger.info(f"Responding to {message_type} with a {response_type} message: {response_payload}.")
        return response_type, response_payload

    def error_response(self, message_type, error_code, error_description):
        if message_type == 'oadrCreatePartyRegistration':
            response_type = 'oadrCreatedPartyRegistration'
        if message_type == 'oadrRequestEvent':
            response_type = 'oadrDistributeEvent'
        else:
            response_type = 'oadrResponse'
        response_payload = {'response': {'response_code': error_code,
                                         'response_description': error_description}}
        return response_type, response_payload



# SPDX-License-Identifier: Apache-2.0

# Copyright 2020 Contributors to OpenLEADR

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# flake8: noqa

import logging
from .client import OpenADRClient
from .server import OpenADRServer


def enable_default_logging(level=logging.INFO):
    """
    Turn on logging to stdout.
    :param level integer: The logging level you wish to use.
                          Defaults to logging.INFO.
    """
    import sys
    import logging
    logger = logging.getLogger('openleadr')
    handler_names = [handler.name for handler in logger.handlers]
    if 'openleadr_default_handler' not in handler_names:
        logger.setLevel(level)
        logging_handler = logging.StreamHandler(stream=sys.stdout)
        logging_handler.set_name('openleadr_default_handler')
        logging_handler.setLevel(logging.DEBUG)
        logger.addHandler(logging_handler)



# SPDX-License-Identifier: Apache-2.0

# Copyright 2020 Contributors to OpenLEADR

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from aiohttp import web
from openleadr.service import EventService, PollService, RegistrationService, ReportService, \
                              OptService, VTNService
from openleadr.messaging import create_message
from openleadr import objects
from openleadr import utils
from functools import partial
from datetime import datetime, timedelta, timezone
from collections import deque
import asyncio
import inspect
import logging
import ssl
import re
import sys
logger = logging.getLogger('openleadr')


class OpenADRServer:
    _MAP = {'on_created_event': 'event_service',
            'on_request_event': 'event_service',

            'on_register_report': 'report_service',
            'on_create_report': 'report_service',
            'on_created_report': 'report_service',
            'on_request_report': 'report_service',
            'on_update_report': 'report_service',

            'on_poll': 'poll_service',

            'on_query_registration': 'registration_service',
            'on_create_party_registration': 'registration_service',
            'on_cancel_party_registration': 'registration_service'}

    def __init__(self, vtn_id, cert=None, key=None, passphrase=None, fingerprint_lookup=None,
                 show_fingerprint=True, http_port=8080, http_host='127.0.0.1', http_cert=None,
                 http_key=None, http_key_passphrase=None, http_path_prefix='/OpenADR2/Simple/2.0b',
                 requested_poll_freq=timedelta(seconds=10), http_ca_file=None):
        """
        Create a new OpenADR VTN (Server).

        :param str vtn_id: An identifier string for this VTN. This is how you identify yourself
                              to the VENs that talk to you.
        :param str cert: Path to the PEM-formatted certificate file that is used to sign outgoing
                            messages
        :param str key: Path to the PEM-formatted private key file that is used to sign outgoing
                           messages
        :param str passphrase: The passphrase used to decrypt the private key file
        :param callable fingerprint_lookup: A callable that receives a ven_id and should return the
                                            registered fingerprint for that VEN. You should receive
                                            these fingerprints outside of OpenADR and configure them
                                            manually.
        :param bool show_fingerprint: Whether to print the fingerprint to your stdout on startup.
                                         Defaults to True.
        :param int http_port: The port that the web server is exposed on (default: 8080)
        :param str http_host: The host or IP address to bind the server to (default: 127.0.0.1).
        :param str http_cert: The path to the PEM certificate for securing HTTP traffic.
        :param str http_key: The path to the PEM private key for securing HTTP traffic.
        :param str http_ca_file: The path to the CA-file that client certificates are checked against.
        :param str http_key_passphrase: The passphrase for the HTTP private key.
        """
        # Set up the message queues
        self.message_queues = {}

        self.app = web.Application()
        self.services = {'event_service': EventService(vtn_id, message_queues=self.message_queues),
                         'report_service': ReportService(vtn_id, message_queues=self.message_queues),
                         'poll_service': PollService(vtn_id, message_queues=self.message_queues),
                         'opt_service': OptService(vtn_id),
                         'registration_service': RegistrationService(vtn_id,
                                                                     poll_freq=requested_poll_freq)}
        if http_path_prefix[-1] == "/":
            http_path_prefix = http_path_prefix[:-1]
        self.app.add_routes([web.post(f"{http_path_prefix}/{s.__service_name__}", s.handler)
                             for s in self.services.values()])
        self.http_port = http_port
        self.http_host = http_host
        self.http_path_prefix = http_path_prefix

        # Create SSL context for running the server
        if http_cert and http_key:
            self.ssl_context = ssl.create_default_context(cafile=http_ca_file,
                                                          purpose=ssl.Purpose.CLIENT_AUTH)
            self.ssl_context.verify_mode = ssl.CERT_REQUIRED
            self.ssl_context.load_cert_chain(http_cert, http_key, http_key_passphrase)
        else:
            self.ssl_context = None

        # Configure message signing
        if cert and key:
            with open(cert, "rb") as file:
                cert = file.read()
            with open(key, "rb") as file:
                key = file.read()
            if show_fingerprint:
                print("")
                print("*" * 80)
                print("Your VTN Certificate Fingerprint is "
                      f"{utils.certificate_fingerprint(cert)}".center(80))
                print("Please deliver this fingerprint to the VENs that connect to you.".center(80))
                print("You do not need to keep this a secret.".center(80))
                print("*" * 80)
                print("")
        VTNService._create_message = partial(create_message, cert=cert, key=key,
                                             passphrase=passphrase)
        VTNService.fingerprint_lookup = staticmethod(fingerprint_lookup)
        self.__setattr__ = self.add_handler

    async def run(self):
        """
        Starts the server in an already-running asyncio loop.
        """
        self.app_runner = web.AppRunner(self.app)
        await self.app_runner.setup()
        site = web.TCPSite(self.app_runner,
                           port=self.http_port,
                           host=self.http_host,
                           ssl_context=self.ssl_context)
        await site.start()
        protocol = 'https' if self.ssl_context else 'http'
        print("")
        print("*" * 80)
        print("Your VTN Server is now running at ".center(80))
        print(f"{protocol}://{self.http_host}:{self.http_port}{self.http_path_prefix}".center(80))
        print("*" * 80)
        print("")

    async def run_async(self):
        await self.run()

    async def stop(self):
        if sys.version_info.minor >= 8:
            delayed_call_tasks = [task for task in asyncio.all_tasks()
                                  if task.get_name().startswith('DelayedCall')]
            for task in delayed_call_tasks:
                task.cancel()
        await self.app_runner.cleanup()

    def add_event(self, ven_id, signal_name, signal_type, intervals, callback=None, event_id=None,
                  targets=None, targets_by_type=None, target=None, response_required='always',
                  market_context="oadr://unknown.context", notification_period=None,
                  ramp_up_period=None, recovery_period=None):
        """
        Convenience method to add an event with a single signal.

        :param str ven_id: The ven_id to whom this event must be delivered.
        :param str signal_name: The OpenADR name of the signal; one of openleadr.objects.SIGNAL_NAME
        :param str signal_type: The OpenADR type of the signal; one of openleadr.objects.SIGNAL_TYPE
        :param str intervals: A list of intervals with a dtstart, duration and payload member.
        :param str callback: A callback function for when your event has been accepted (optIn) or refused (optOut).
        :param list targets: A list of Targets that this Event applies to.
        :param target: A single target for this event.
        :param dict targets_by_type: A dict of targets, grouped by type.
        :param str market_context: A URI for the DR program that this event belongs to.
        :param timedelta notification_period: The Notification period for the Event's Active Period.
        :param timedelta ramp_up_period: The Ramp Up period for the Event's Active Period.
        :param timedelta recovery_period: The Recovery period for the Event's Active Period.

        If you don't provide a target using any of the three arguments, the target will be set to the given ven_id.
        """
        if self.services['event_service'].polling_method == 'external':
            logger.error("You cannot use the add_event method after you assign your own on_poll "
                         "handler. If you use your own on_poll handler, you are responsible for "
                         "delivering events from that handler. If you want to use OpenLEADRs "
                         "message queuing system, you should not assign an on_poll handler. "
                         "Your Event will NOT be added.")
            return
        if not re.match(r"^(([^:/?#]+):)?(//([^/?#]*))?([^?#]*)(\?([^#]*))?(#(.*))?", market_context):
            raise ValueError("The Market Context must be a valid URI.")
        event_id = event_id or utils.generate_id()

        if response_required not in ('always', 'never'):
            raise ValueError("'response_required' should be either 'always' or 'never'; "
                             f"you provided '{response_required}'.")

        # Figure out the target for this Event
        if target is None and targets is None and targets_by_type is None:
            targets = [{'ven_id': ven_id}]
        elif target is not None:
            targets = [target]
        elif targets_by_type is not None:
            targets = utils.ungroup_targets_by_type(targets_by_type)
        if not isinstance(targets, list):
            targets = [targets]

        event_descriptor = objects.EventDescriptor(event_id=event_id,
                                                   modification_number=0,
                                                   market_context=market_context,
                                                   event_status="far",
                                                   created_date_time=datetime.now(timezone.utc))
        event_signal = objects.EventSignal(intervals=intervals,
                                           signal_name=signal_name,
                                           signal_type=signal_type,
                                           signal_id=utils.generate_id(),
                                           targets=targets)

        # Make sure the intervals carry timezone-aware timestamps
        for interval in intervals:
            if utils.getmember(interval, 'dtstart').tzinfo is None:
                utils.setmember(interval, 'dtstart',
                                utils.getmember(interval, 'dtstart').astimezone(timezone.utc))
                logger.warning("You supplied a naive datetime object to your interval's dtstart. "
                               "This will be interpreted as a timestamp in your local timezone "
                               "and then converted to UTC before sending. Please supply timezone-"
                               "aware timestamps like datetime.datetime.new(timezone.utc) or "
                               "datetime.datetime(..., tzinfo=datetime.timezone.utc)")
        active_period = utils.get_active_period_from_intervals(intervals, False)
        active_period.ramp_up_period = ramp_up_period
        active_period.notification_period = notification_period
        active_period.recovery_period = recovery_period
        event = objects.Event(active_period=active_period,
                              event_descriptor=event_descriptor,
                              event_signals=[event_signal],
                              targets=targets,
                              response_required=response_required)
        self.add_raw_event(ven_id=ven_id, event=event, callback=callback)
        return event_id

    def add_raw_event(self, ven_id, event, callback=None):
        """
        Add a new event to the queue for a specific VEN.
        :param str ven_id: The ven_id to which this event should be distributed.
        :param dict event: The event (as a dict or as a objects.Event instance)
                           that contains the event details.
        :param callable callback: A callback that will receive the opt status for this event.
                                  This callback receives ven_id, event_id, opt_type as its arguments.
        """
        if utils.getmember(event, 'response_required') == 'always':
            if callback is None:
                logger.warning("You did not provide a 'callback', which means you won't know if the "
                               "VEN will opt in or opt out of your event. You should consider adding "
                               "a callback for this.")
            elif not asyncio.isfuture(callback):
                args = inspect.signature(callback).parameters
                if not all(['ven_id' in args, 'event_id' in args, 'opt_type' in args]):
                    raise ValueError("The 'callback' must have at least the following parameters: "
                                     "'ven_id' (str), 'event_id' (str), 'opt_type' (str). Please fix "
                                     "your 'callback' handler.")

        if ven_id not in self.message_queues:
            self.message_queues[ven_id] = deque()
        event_id = utils.getmember(utils.getmember(event, 'event_descriptor'), 'event_id')
        self.message_queues[ven_id].append(event)
        if callback is not None:
            self.services['event_service'].pending_events[event_id] = (event, callback)
        if utils.getmember(event, 'response_required') == 'never':
            self.services['event_service'].schedule_event_updates(ven_id, event)
        return event_id

    def add_handler(self, name, func):
        """
        Add a handler to the OpenADRServer.

        :param str name: The name for this handler. Should be one of: on_created_event,
                            on_request_event, on_register_report, on_create_report,
                            on_created_report, on_request_report, on_update_report, on_poll,
                            on_query_registration, on_create_party_registration,
                            on_cancel_party_registration.
        :param callable func: A function or coroutine that handles this type of occurrence.
                              It receives the message, and should return the contents of a response.
        """
        logger.debug(f"Adding handler: {name} {func}")
        if name in self._MAP:
            setattr(self.services[self._MAP[name]], name, func)
            if name == 'on_poll':
                self.services['poll_service'].polling_method = 'external'
                self.services['event_service'].polling_method = 'external'
        else:
            raise NameError(f"""Unknown handler '{name}'. """
                            f"""Correct handler names are: '{"', '".join(self._MAP.keys())}'.""")
